import * as FS from "fs";
//import * as Path from "path";

export class Files {
    
    public static async ReadUTF8(location:string):Promise<string> {
        let path = Files.InterpretPath(location);
        return new Promise<string>((resolve, reject) => {
            FS.readFile(path, 'utf8', function(error, data) {
                if(error != null) { return reject(error); }
                return resolve(data);
            });
        });
    }

    public static async Read(location:string):Promise<Buffer> {
        let path = Files.InterpretPath(location);
        return new Promise<Buffer>((resolve, reject) => {
            FS.readFile(path, function(error, data) {
                if(error != null) { return reject(error); }
                return resolve(data);
            });
        });
    }

    public static async ReadFileToBuffer(location:string):Promise<Buffer> {
        let path = Files.InterpretPath(location);
        return new Promise<Buffer>((resolve, reject) => {
            FS.readFile(path, function(error, data) {
                if(error != null) { return reject(error); }
                return resolve(data);
            });
        });
    }

    public static async Write(location:string, contents:(string)|Buffer):Promise<void> {
        let path = Files.InterpretPath(location);
        await Files.AssertFilePath(path);
        return new Promise<void>((resolve, reject) => {
            FS.writeFile(path, contents, function(error) {
                if(error != null) { return reject(error); }
                return resolve();
            });
        });
    }

    public static async Delete(location:string):Promise<void> {
        let exists = await Files.Exists(location);
        if(exists == false) { return; }

        let lstat = await Files.Stat(location);
        if(lstat.isDirectory()) {

            let files = await Files.ReadDir(location);
            for(let i = 0; i < files.length; i++) {
                await Files.Delete(location + "/" + files[i]);
            }

            let path = Files.InterpretPath(location);
            return new Promise<void>((resolve, reject) => {
                FS.rmdir(path, function(error) {
                    if(error != null) { return reject(error); }
                    return resolve();
                });
            });

        } else {
            let path = Files.InterpretPath(location);
            return new Promise<void>((resolve, reject) => {
                FS.unlink(path, function(error) {
                    if(error != null) { return reject(error); }
                    return resolve();
                });
            });
        }
    }

    public static async Exists(location:string):Promise<boolean> {
        let path = Files.InterpretPath(location);
        return new Promise<boolean>((resolve, reject) => {
            FS.exists(path, function(exists) {
                return resolve(exists);
            });
        });
    }

    public static async MkDir(location:string):Promise<void> {
        let path = Files.InterpretPath(location);
        return new Promise<void>((resolve, reject) => {
            FS.mkdir(path, function(err) {
                if(err != null) { return reject(err); }
                return resolve();
            });
        });
    }

    public static async AssertFilePath(file_path:string):Promise<void> {

        let tokens = file_path.split("/");

        // drop the file token
        tokens = tokens.slice(0, tokens.length - 1);

        let idx = 1;
        let finished = (idx >= tokens.length);
        while(finished == false) {
            let current_tokens = tokens.slice(0, idx + 1);
            let current_path = current_tokens.join("/");

            let exists = await Files.Exists(current_path);
            if(!exists) {
                await Files.MkDir(current_path);
            }

            idx ++;
            finished = (idx >= tokens.length);
        }
    }

    public static async Stat(location:string):Promise<FS.Stats> {
        let path = Files.InterpretPath(location);
        return new Promise<FS.Stats>((resolve, reject) => {
            FS.stat(path, function(error, stat) {
                if(error != null) { return reject(error); }
                return resolve(stat);
            });
        });
    }

    public static async ReadDir(location:string):Promise<Array<string>> {
        let path = Files.InterpretPath(location);
        return new Promise<Array<string>>((resolve, reject) => {
            FS.readdir(path, function(error, data) {
                if(error != null) { return reject(error); }
                return resolve(data);
            });
        });
    }

    public static async ReadDirRecursively(location:string, relative_location?:string, exclude?:string):Promise<Array<ILocationStat>> {

        if(exclude != null && location.indexOf(exclude) >= 0) { return []; }

        if(relative_location == null) { relative_location = ""; }
        else { relative_location = relative_location + "/"; }

        let files = await Files.ReadDir(location);
        let result:Array<ILocationStat> = [];
        for(let i = 0; i < files.length; i++) {
            let stat = await Files.Stat(location + "/" + files[i]);
            if(stat.isFile()) {
                result.push({
                    path: relative_location + files[i],
                    stat: stat
                });
                continue;
            }
            if(stat.isDirectory()) {
                let dir_files = await Files.ReadDirRecursively(location + "/" + files[i], relative_location + files[i], exclude);
                result = result.concat(dir_files);
                continue;
            }
        }

        return result;
    }

    public static ReadStream(location:string):FS.ReadStream {
        let path = Files.InterpretPath(location);
        return FS.createReadStream(path)
    }

    public static async Copy(src:string, dst:string):Promise<void> {

        let contents = await Files.Read(src);
        await Files.Write(dst, contents);
    }

    public static InterpretPath(path:string):string {

        if(!path.startsWith("/")) {
            path = this.RelativePath() + path;
        }

        let tokens = path.split("/");
        let result:Array<string> = [];
        for(let i = 0; i < tokens.length; i++) {
            if(tokens[i] == "..") { result.pop(); }
            else { result.push(tokens[i]); }
        }
        return result.join("/");
    }

    private static RelativePath():string {
        return __dirname.substring(0, __dirname.lastIndexOf("/") + 1);
    }
}

export interface ILocationStat {
    path: string;
    stat: FS.Stats;
}