const SERVER_LOCATION:string = "localhost:8080";

import * as Busboy from "busboy";
import * as HTTP from "http";
import * as REQUEST from "request";
import * as FormData from "form-data";
import * as QueryString from "querystring";

import * as Utilities from "../../utilities/All";

export class ServerRequest {

    public static async Get(path:string, query:any, localRequest:HTTP.IncomingMessage, localResponse?:HTTP.ServerResponse):Promise<ServerResponse> {

        var sendPath = "/service" + path;
        if(query != null) {
            sendPath = sendPath + "?" + QueryString.stringify(query);
        }

        // send the headers of the local request along (we're doing this proxy style)
        return new Promise<ServerResponse>((resolve, reject) => {

            let content = null;
            var request = HTTP.request(
                {
                    method: "GET",
                    headers: localRequest.headers,
                    hostname: SERVER_LOCATION.split(":")[0],
                    port: Number.parseInt(SERVER_LOCATION.split(":")[1]),
                    path: sendPath
                }, function(response:HTTP.IncomingMessage) {

                    response.setEncoding('utf8');
                    response.on('data', (chunk) => {
                        if(content == null) { content = ""; }
                        content = content + chunk;
                    });
                    response.on('end', () => {

                        resolve({
                            request: response,
                            content: content
                        });
                    });
                }
            );
            request.on("error", function(err:Error) {
                reject(err);
            });
            request.end();
        });
    }

    public static async GetWithHeaders(path:string, query:any, headers:any):Promise<ServerResponse> {

        var sendPath = "/service" + path;
        if(query != null) {
            sendPath = sendPath + "?" + QueryString.stringify(query);
        }

        // send the headers of the local request along (we're doing this proxy style)
        return new Promise<ServerResponse>((resolve, reject) => {

            let content = null;
            var request = HTTP.request(
                {
                    method: "GET",
                    headers: headers,
                    hostname: SERVER_LOCATION.split(":")[0],
                    port: Number.parseInt(SERVER_LOCATION.split(":")[1]),
                    path: sendPath
                }, function(response:HTTP.IncomingMessage) {

                    response.setEncoding('utf8');
                    response.on('data', (chunk) => {
                        if(content == null) { content = ""; }
                        content = content + chunk;
                    });
                    response.on('end', () => {

                        resolve({
                            request: response,
                            content: content
                        });
                    });
                }
            );
            request.on("error", function(err:Error) {
                reject(err);
            });
            request.end();
        });
    }

    public static async PostWithHeaders(path:string, query:any, headers:any, json:any, files:FileUploads):Promise<ServerResponse> {

        // form sender
        var formData = new FormData();

        // parse form data using the keys
        formData.append("data", JSON.stringify(json));
        let keys = Object.keys(files);
        for(let i = 0; i < keys.length; i++) {
            let key = keys[i];
            let value = files[key];
            formData.append(value.name, value.data, { filename: value.filename});
        }

        var customHeaders:any = formData.getHeaders();
        if(customHeaders == null) {
            customHeaders = {};
        }

        if(headers != null) {
            let keys = Object.keys(headers);
            for(let i = 0; i < keys.length; i++) {
                let key = keys[i];
                let value = headers[key];
                if(key.trim().toLowerCase().startsWith("content-type")) {
                    continue;
                }
                customHeaders[key] = value;
            }
        }

        /*

        let options = {
            headers: headers
        };
        */

        var sendPath = "/service" + path;
        if(query != null) {
            sendPath = sendPath + "?" + QueryString.stringify(query);
        }

        return new Promise<ServerResponse>((resolve, reject) => {

            let content = null;

            /*
            let request = REQUEST.post("http://" + SERVER_LOCATION + sendPath, options, function(err, response, body){ 
                if(err != null) { return reject(err); }
                resolve(response);
            });
            request.form(form);
            */
            var request = HTTP.request({
                method: 'post',
                hostname: SERVER_LOCATION.split(":")[0],
                port: Number.parseInt(SERVER_LOCATION.split(":")[1]),
                path: sendPath,
                headers: customHeaders
            }, function(response:HTTP.IncomingMessage) {

                    response.setEncoding('utf8');
                    response.on('data', (chunk) => {
                        if(content == null) { content = ""; }
                        content = content + chunk;
                    });
                    response.on('end', () => {

                        resolve({
                            request: response,
                            content: content
                        });
                    });
                }
            );
            request.on("error", function(err:Error) {
                reject(err);
            });
            formData.pipe(request);
            //request.end();
        });
    }

    public static async Proxy(from:ServerResponse, to:HTTP.ServerResponse):Promise<void> {

        // copy cookies
        Utilities.Web.CopyCookies(from.request, to);

        to.statusCode = from.request.statusCode;
        to.statusMessage = from.request.statusMessage;
        to.end(from.content);
    }

    public static async ProxyGetRequest(url:string, localRequest:HTTP.IncomingMessage, localResponse:HTTP.ServerResponse):Promise<void> {
        var queryParameters = await Utilities.Web.GetRequestParameter(localRequest);
        var serverResponse = await ServerRequest.Get(url, {data:JSON.stringify(queryParameters.fields)}, localRequest);
        await ServerRequest.Proxy(serverResponse, localResponse);
    }

    public static async ProxyPostRequest(url:string, localRequest:HTTP.IncomingMessage, localResponse:HTTP.ServerResponse):Promise<void> {
        var queryParameters = await Utilities.Web.GetRequestParameter(localRequest);
        var serverResponse = await ServerRequest.Get(url, { data:JSON.stringify(queryParameters.fields) }, localRequest);
        await ServerRequest.Proxy(serverResponse, localResponse);
    }

    public static async ParseFormData(request:HTTP.IncomingMessage):Promise<FieldsAndFiles> {
        return new Promise<FieldsAndFiles>((resolve, reject) => {

            let fields = {};
            let files = {};

            var busboy = new Busboy({ headers: request.headers });
            busboy.on('file', function(fieldname, file, filename, encoding, mimetype) {
                let fileBuffer:Buffer = new Buffer("");
                file.on('data', function(data) {
                    fileBuffer = Buffer.concat([fileBuffer, data]);
                });
                file.on('end', function() {
                    files[fieldname] = { path: null, data: fileBuffer, originalFilename: filename };
                });
            });
            busboy.on('field', function(fieldname, val, fieldnameTruncated, valTruncated, encoding, mimetype) {
                fields[fieldname] = val;
            });
            busboy.on('finish', function() {
                resolve({
                    fields: fields,
                    files: files
                });
            });
            request.pipe(busboy);
        });
    }

    public static async ParseJsonData(request:HTTP.IncomingMessage):Promise<any> {

        let content = null;

        return new Promise<any>((resolve, reject) => {
            request.on('data', function(chunk) {
                if(content == null) { content = ""; }
                content = content + chunk;
            });
            request.on('end', function() {
                resolve(JSON.parse(content));
            });
            request.on('error', function(err) {
                reject(err);
            })

        });
    }
}

interface ServerResponse {
    request: HTTP.IncomingMessage;
    content: any;
}

export interface FieldsAndFiles {
    fields: any;
    files: any;
}

export interface FileUploads {
    [name:string]:FileUpload;
}

interface FileUpload {
    name: string;
    data: Buffer;
    filename: string;
}