import * as HTTP from "http";
import * as Utilities from "../../../utilities/All";
import { ServerRequest, FileUploads } from "../ServerRequest";
//import * as MultiParty from "multiparty";
import * as Crypto from "crypto";
//import * as Archiver from "archiver";

export class Build {

    private static HashSets:{[module_id:string]:HashSet} = {};
    private static ModifiedTimestamps:{[module_id:string]:ModifiedTimestamps} = {};

    public static async BuildLocal(request:HTTP.IncomingMessage, response:HTTP.ServerResponse):Promise<void> {

        //var queryParameters = await ServerRequest.ParseFormData(request);
        //var queryParameters = await Utilities.Web.GetRequestParameter(request);
        //var fields = queryParameters.fields;
        var fields = await ServerRequest.ParseJsonData(request);
        
        let source:any = null;
        let module:any = null;
        let hashSet:any = null;

        if(fields.module != null && fields.source != null) {
            module = fields.module;
            source = fields.source;
            hashSet = fields.hashSet;
        }

        let meta = await Build.GetMeta(module, source, hashSet);

        if(meta.project_files.length > 0 || meta.missing_files.length > 0) {

            let files:Array<string> = meta.project_files.slice();
            await Utilities.Files.Write(source.Location + "/.framework.sh.meta", JSON.stringify(meta));
            files.push(".framework.sh.meta");
            await Utilities.Archive.ArchiveFilesToResponse(source.Location, files, response);
            await Utilities.Files.Delete(source.Location + "/.framework.sh.meta");
            return;
        }

        response.writeHead(200, {
            'Content-Type': 'application/zip',
            'Content-Length': 0
        });

        return new Promise<void>((resolve, reject) => {
            response.end(function() {
                resolve();
            });
            response.on('error', function(err) {
                reject(err);
            });
        });
    }

    public static async BuildUpdate(request:HTTP.IncomingMessage, response:HTTP.ServerResponse):Promise<void> {

        //var queryParameters = await ServerRequest.ParseJsonData(request);
        //var queryParameters = await Utilities.Web.GetRequestParameter(request);
        //var fields = queryParameters.fields;
        var fields = await ServerRequest.ParseJsonData(request);

        let source:any = fields.source;
        let module:any = fields.module;
        let build:IBuild = fields.build;

        if(build.FrameworkDTS != null) {
            let frameworkDTS:string = null;
            try {
                frameworkDTS = await Utilities.Files.ReadUTF8(source.Location + "/framework.d.ts");
            } catch(err) { }
            if(frameworkDTS == null || frameworkDTS.trim() != build.FrameworkDTS.trim()) {
                await Utilities.Files.Write(source.Location + "/framework.d.ts", build.FrameworkDTS);
            }
        }

        let json:string = JSON.stringify(null);
        response.writeHead(200, {
            'Content-Type': 'application/json',
            'Content-Length': json.length
        });

        return new Promise<void>((resolve, reject) => {
            response.write(json);
            response.end(function() {
                resolve();
            });
            response.on('error', function(err) {
                reject(err);
            });
        });
    }

    private static async GetMeta(module:any, source:any, remoteHashSet:HashSet):Promise<FrameworkShMeta> {

        var location = source.Location;
        var fileArray = await Utilities.Files.ReadDirRecursively(location, null, "node_modules");
        // get a copy of the local hash set
        // if a local one does not exist then read / hash the files
        let localHashSet = Build.HashSets[module.id];
        let modifiedTimestamps = Build.ModifiedTimestamps[module.id];
        let localHashSetNew:boolean = false;
        if(localHashSet == null) {
            localHashSetNew = true;
            localHashSet = {
                Paths: [],
                Hashes: {}
            };
            modifiedTimestamps = {
                Timestamps: {}
            };
            let promises:Array<Promise<any>> = new Array(fileArray.length);
            for(let i = 0; i < fileArray.length; i++) {
                let path = fileArray[i].path;

                // timestamp
                modifiedTimestamps.Timestamps[path] = fileArray[i].stat.mtime;

                let promise = async function() {
                    let hash = await Build.HashFile(location + "/" + path);
                    localHashSet.Paths.push(path);
                    localHashSet.Hashes[path] = hash;
                };
                promises[i] = promise();
                /*
                // hash
                let contents = await Utilities.Files.ReadFileToBuffer(location + "/" + path);
                let hash = RevHash(contents);
                localHashSet.Paths.push(path);
                localHashSet.Hashes[path] = hash;

                // timestamp
                modifiedTimestamps.Timestamps[path] = fileArray[i].stat.mtime;
                */
            }
            await Promise.all(promises);
            Build.HashSets[module.id] = localHashSet;
            Build.ModifiedTimestamps[module.id] = modifiedTimestamps;
        }

        // by looping over the local files
        // we can detect new / modified files
        let newModifiedFiles:Array<string> = [];
        let fileMap:{[path:string]:number} = {};
        let changeDetectedPromises:Array<Promise<any>> = [];
        for(let i = 0; i < fileArray.length; i++) {
            
            let file = fileArray[i];
            let timestamp = modifiedTimestamps.Timestamps[file.path];
            let localHash = localHashSet.Hashes[file.path];
            let remoteHash = remoteHashSet == null ? null : remoteHashSet.Hashes[file.path];

            fileMap[file.path] = -1;

            // actively scrub .js and .js.map files
            let isTs:boolean = file.path.endsWith(".ts");
            let isJs:boolean = file.path.endsWith(".js");
            let isMap:boolean = file.path.endsWith(".js.map");
            if(isTs || isJs || isMap) {
                
                let nameOnly = file.path;
                if(isTs) { nameOnly = nameOnly.substring(0, nameOnly.lastIndexOf(".ts")); }
                else if(isJs) { nameOnly = nameOnly.substring(0, nameOnly.lastIndexOf(".js")); }
                else if(isMap) { nameOnly = nameOnly.substring(0, nameOnly.lastIndexOf(".js.map")); }

                // is TS file (scrub JS / MAP and continue)
                if(isTs) {
                    let jsIdx = fileMap[nameOnly + ".js"];
                    let mapIdx = fileMap[nameOnly + ".js.map"];

                    if(jsIdx != null && jsIdx >= 0) { newModifiedFiles[jsIdx] = null; }
                    if(mapIdx != null && mapIdx >= 0) { newModifiedFiles[mapIdx] = null; }

                    if(localHashSet.Hashes[nameOnly + ".js"] != null) { 
                        delete localHashSet.Hashes[nameOnly + ".js"]; 
                        delete modifiedTimestamps.Timestamps[nameOnly + ".js"];
                    }

                    if(localHash[nameOnly + ".js.map"] != null) {
                        delete localHashSet.Hashes[nameOnly + ".js.map"]; 
                        delete modifiedTimestamps.Timestamps[nameOnly + ".js.map"];
                    }
                }

                // is JS file (scrub MAP and continue)
                else if(isJs) {
                    let tsIdx = fileMap[nameOnly + ".ts"];
                    let tsFileFound = tsIdx != null;

                    if(tsFileFound) {
                        let mapIdx = fileMap[nameOnly + ".js.map"];
                        if(mapIdx != null && mapIdx >= 0) { newModifiedFiles[mapIdx] = null; }
                        if(localHash[nameOnly + ".js.map"] != null) {
                            delete localHashSet.Hashes[nameOnly + ".js.map"]; 
                            delete modifiedTimestamps.Timestamps[nameOnly + ".js.map"];
                        }
                        continue;
                    }
                }

                // is MAP file (scrub JS and continue)
                else if(isMap) {
                    let tsIdx = fileMap[nameOnly + ".ts"];
                    let tsFileFound = tsIdx != null;

                    if(tsFileFound) {
                        let jsIdx = fileMap[nameOnly + ".js"];
                        if(jsIdx != null && jsIdx >= 0) { newModifiedFiles[jsIdx] = null; }
                        if(localHash[nameOnly + ".js"] != null) {
                            delete localHashSet.Hashes[nameOnly + ".js"]; 
                            delete modifiedTimestamps.Timestamps[nameOnly + ".js"];
                        }
                        continue;
                    }
                }
            }

            // this means the file is new || modified
            if(remoteHash == null || localHash != remoteHash || timestamp == null || file.stat.mtime == null || timestamp.getTime() < file.stat.mtime.getTime()) {
                // update mtime / hash
                modifiedTimestamps.Timestamps[file.path] = file.stat.mtime;
                let promise = async function() {
                    if(localHashSetNew == false) {
                        //let contents = await Utilities.Files.ReadFileToBuffer(location + "/" + file.path);
                        //let hash = RevHash(contents);
                        let hash = await Build.HashFile(location + "/" + file.path);
                        localHashSet.Hashes[file.path] = hash;
                    }
                    newModifiedFiles.push(file.path);
                    fileMap[file.path] = newModifiedFiles.length - 1;
                };
                changeDetectedPromises.push(promise());
            }
        }

        // await the change detection (in parallel)
        await Promise.all(changeDetectedPromises);

        // compress the newModifiedFiles
        let tempResult:Array<string> = [];
        for(let i = 0; i < newModifiedFiles.length; i++) {
            let file = newModifiedFiles[i];
            if(file == null) { continue; }
            tempResult.push(file);
        }
        newModifiedFiles = tempResult;

        // by looping over the remote files
        // we can detect missing files (these must be deleted remotely)
        let missingFiles:Array<string> = [];
        if(remoteHashSet != null) {
            for(let i = 0; i < remoteHashSet.Paths.length; i++) {
                let path = remoteHashSet.Paths[i];
                if(localHashSet.Hashes[path] == null) {
                    console.log("$$missing: " + path);
                    missingFiles.push(path);
                }
            }
        }

        return {
            project_files: newModifiedFiles,
            missing_files: missingFiles,
            hashes: localHashSet.Hashes
        };
    }

    public static async HashFile(location:string):Promise<string> {
        let stream = Utilities.Files.ReadStream(location);
        let hash = Crypto.createHash("md5");
        return new Promise<string>((resolve, reject) => {
            stream
                .on("data", function(chunk) { hash.update(chunk); })
                .on("end", function() { resolve(hash.digest("hex").slice(0, 10)); })
                .on("error", function(err) { reject(err); })
        });
    }
}

interface FrameworkShMeta {
    project_files: Array<string>;
    missing_files: Array<string>;
    hashes: {[path:string]:string};
}

interface Module {
    id: string;
    Name: string;
    Location: string;
}

interface HashSet {
    Paths:Array<string>;
    Hashes:{[path:string]:string;};
}

interface ModifiedTimestamps {
    Timestamps:{[path:string]:Date;};
}

interface IBuild {
    CreatedOn: Date;
    IsBuildBusy: boolean;
    StartedOn: Date;
    IsBuildSuccess: boolean;
    SuccessOn: Date;
    IsBuildFailed: boolean;
    FailedOn: Date;
    FrameworkDTS: Types.Text;
    DataSourcesJSON: Types.Text;
    DataSourcesExportsJSON: Types.Text;
    QueuesJSON: Types.Text;
    ServicesJSON: Types.Text;
    ExternalWebServicesJSON: Types.Text;
    SymbolReferencesJSON: Types.Text;
    UserFK: string;
    ModuleFK: string;
}